
/* This file is part of indexlib.
 * Copyright (C) 2005 Lus Pedro Coelho <luis@luispedro.org>
 *
 * Indexlib is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation and available as file
 * GPL_V2 which is distributed along with indexlib.
 * 
 * Indexlib is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of this program with any edition of
 * the Qt library by Trolltech AS, Norway (or with modified versions
 * of Qt that use the same license as Qt), and distribute linked
 * combinations including the two.  You must obey the GNU General
 * Public License in all respects for all of the code used other than
 * Qt.  If you modify this file, you may extend this exception to
 * your version of the file, but you are not obligated to do so.  If
 * you do not wish to do so, delete this exception statement from
 * your version.
 */

#include "character.h"
#include <cctype>
#include <cstring>


bool valid_character( character c ) {
	return c == stop_character || c == dollar_character || c >= 'A' && c <= 'Z';
}

/* Valid Characters
 * A ... Z 
 * 0 ... 9
 */

character convert( char c ) {
	if ( c >= 'A' && c <= 'Z' ) return c;
	if ( c >= 'a' && c <= 'z' ) return std::toupper( c );
	switch ( c ) {

		// No, this is not a joke.
		// We are saving space by losing some precision.
		//
		// The reasoning is that if you look for a number it
		// will, after translation, look completelly different from a
		// normal text string so confusion is not probable.
		//
		// For example, if you look for the number 1024, then you will be
		// looking for the string JZQX which will probably not appear elsewhere.
		//
		// If you are wondering about the choice of letters, these
		// are the least frequently used characters in the English language
		// according to
		//    http://en.wikipedia.org/wiki/Letter_frequencies
		case '0': return 'Z';
		case '1': return 'J';
		case '2': return 'Q';
		case '3': return 'K';
		case '4': return 'X';
		case '5': return 'B';
		case '6': return 'V';
		case '7': return 'W';
		case '8': return 'G';
		case '9': return 'Y';

		case '':
		case '':
		case '':
		case '':
		case '':
		case '':
		case '':
		case '':
		case '':
		case '':
			return 'A';

		case '':
		case '':
			return 'C';

		case '':
		case '':
		case '':
		case '':
			return 'E';

		case '':
		case '':
		case '':
		case '':
		case '':
		case '':
			return 'O';

		case '':
		case '':
			return 'I';

		case ' ':
		case '\n':
		case '\t':

		case '?':
		case '!':

		case '.':
		case ',':
		case ';':
		case ':':
		case '(':
		case ')':
		case '[':
		case ']':
		case '{':
		case '}':
		case '-':
		case '_':
		case '"':
		case '\'':
			return stop_character;
	}
	return nul_character;
}

character_string::character_string() {
}

character_string::character_string( const character* original ):
	data_( original, original + char_len( original ) +  1 )
{
}

character_string::character_string( const char* original ):
	data_( std::strlen( original ) + 1 )
{
	character* out = &data_.front();
	bool last_stop = false;
	while ( *original ) {
		if ( convert( *original ) ) {
			if ( convert( *original ) == stop_character ) {
				if ( !last_stop ) *out++ = convert( *original );
				last_stop = true;
			} else {
				last_stop = false;
				*out++ = convert( *original );
			}
		}
		++original;
	}
	*out++ = nul_character;
	data_.resize( out - &data_.front() );
}

character_string::character_string( std::string str )
{
	*this = character_string( str.c_str() );
}

character_string::character_string( const character_string& other ):
	data_( other.data_ )
{
}

character_string::~character_string() {
}

character_string& character_string::operator = ( const character_string& other ) {
	character_string tmp( other );
	tmp.swap( *this );
	return *this;
}


