#ifndef LPC_CHARACTER_H1102530057_INCLUDE_GUARD_
#define LPC_CHARACTER_H1102530057_INCLUDE_GUARD_

/* This file is part of indexlib.
 * Copyright (C) 2005 Lus Pedro Coelho <luis@luispedro.org>
 *
 * Indexlib is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation and available as file
 * GPL_V2 which is distributed along with indexlib.
 * 
 * Indexlib is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of this program with any edition of
 * the Qt library by Trolltech AS, Norway (or with modified versions
 * of Qt that use the same license as Qt), and distribute linked
 * combinations including the two.  You must obey the GNU General
 * Public License in all respects for all of the code used other than
 * Qt.  If you modify this file, you may extend this exception to
 * your version of the file, but you are not obligated to do so.  If
 * you do not wish to do so, delete this exception statement from
 * your version.
 */

#include <stdint.h>
#include <string.h>
#include <vector>
#include <cassert>
#include <algorithm>
#include <iostream>
#include <boost/format.hpp>

typedef unsigned char character;
const unsigned number_characters = 29;
const unsigned alphabet_size = number_characters;
const character nul_character = '\0';
const character dollar_character = '$';
const character stop_character = '.';

bool valid_character( character );
character convert( char c );
struct character_string {
	public:
		character_string();
		character_string( const char* str );
		character_string( std::string str );

		character_string( const character* str );
		character_string( const character_string& other );
		character_string&
			operator = ( const character_string& other );
		void swap( character_string& other ) {
			data_.swap( other.data_ );
		}
		~character_string();

		character operator[]( unsigned idx ) const { return data_[ idx ]; }
		const character& operator[]( unsigned idx ) { return data_[ idx ]; }
		unsigned size() const { return data_.size() - 1; }
		void resize( unsigned len ) {
			data_.resize( len + 1 );
			data_.back() = nul_character;
		}
		const character* c_str() const {
			return &data_.front();
		}
		const char* cchar_str() const {
			return reinterpret_cast<const char*>( c_str() );
		}
		bool operator == ( const character_string& other ) const;
	private:
		std::vector<character> data_;
};

inline
bool character_string::operator == ( const character_string& other ) const {
	return std::equal( data_.begin(), data_.end(), other.data_.begin() );
}

inline
std::ostream& operator << ( std::ostream& out, const character_string& str ) {
	return out << boost::format( "[%s]%s" ) % str.size() % str.c_str();
}

const character base_character = 'A';
inline uint8_t character2idx( character c ) {
	if ( !c ) return 0;
	assert( valid_character( c ) );
	if ( c == dollar_character ) return number_characters - 1;
	if ( c == stop_character ) return number_characters - 2;
	return c - base_character + 1;
}

inline character idx2character( uint8_t v ) {
	if ( !v ) return nul_character;
	if ( v == number_characters - 1 ) return dollar_character;
	if ( v == number_characters - 2 ) return stop_character;
	return v + base_character - 1;
}

inline unsigned char_len( const character* c ) {
	unsigned res = 0;
	while ( *c++ ) ++res;
	return res;
}

inline unsigned common_chars( const character* s1, const character* s2 ) {
	unsigned res = 0;
	while ( *s1 && *s1++ == *s2++ ) ++res;
	return res;
}
inline
character_string right( const character_string& str, unsigned len ) {
	return character_string( str.c_str() + ( str.size() - len ) );
}

inline
character_string left( const character_string& str, unsigned len ) {
	character_string res = str;
	res.resize( len );
	return res;
}

inline
character_string substr( const character_string& str, unsigned i, unsigned j ) {
	assert( j > i );
	return right( left( str, j ), str.size() - ( j - i ) );
	
}

#endif /* LPC_CHARACTER_H1102530057_INCLUDE_GUARD_ */
