
/* This file is part of indexlib.
 * Copyright (C) 2005 Lus Pedro Coelho <luis@luispedro.org>
 *
 * Indexlib is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation and available as file
 * GPL_V2 which is distributed along with indexlib.
 * 
 * Indexlib is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of this program with any edition of
 * the Qt library by Trolltech AS, Norway (or with modified versions
 * of Qt that use the same license as Qt), and distribute linked
 * combinations including the two.  You must obey the GNU General
 * Public License in all respects for all of the code used other than
 * Qt.  If you modify this file, you may extend this exception to
 * your version of the file, but you are not obligated to do so.  If
 * you do not wish to do so, delete this exception statement from
 * your version.
 */

#include "node.h"
#include "character.h"
#include "bitstream.h"
#include "logfile.h"
#include "stringarray.h"
#include <memory>
#include <string>
#include <boost/format.hpp>

bitstream& operator << ( bitstream& out,  node::nodeptr p ) {
	return out << p.cast_to_uint32();
}


bitstream& operator >> ( bitstream& in, node::nodeptr& p ) {
	uint32_t v;
	in >> v;
	p = node::nodeptr::cast_from_uint32( v );
	return in;
}

namespace {
	memory_manager* node_manager = 0;
}
void node::set_manager( memory_manager* m ) {
	node_manager = m;
}

memory_manager* node::get_manager() {
	return node_manager;
}

const stringarray* node::farbranch::all_strings = 0;


node::nodeptr node::internalnode::map( character c ) {
	if ( type() == nodetypes::ifull ) return internalnode_full( static_cast<base_type&>( *this ) ).do_map( c );
	else if ( type() == nodetypes::ihalf ) return internalnode_half( static_cast<base_type&>( *this ) ).do_map( c );
	assert( 0 );
	return nodeptr();
}


bool node::internalnode::can_add() const {
	if ( type() == nodetypes::ifull ) return true;
	else if ( type() == nodetypes::ihalf ) {
		return internalnode_half( static_cast<const base_type&>( *this ) ).do_can_add();
	}
	assert( 0 );
}


void node::internalnode::add( character c, nodeptr n ) {
	assert( valid_character( c ) );
	assert( !map( c ) );
	substitute( c, n );
}
void node::internalnode::substitute( character c, nodeptr n ) {
	if ( type() == nodetypes::ifull ) {
		internalnode_full that( static_cast<base_type&>( *this ) );
		that.do_substitute( c, n );
	} else if ( type() == nodetypes::ihalf ) {
		internalnode_half that( static_cast<base_type&>( *this ) );
		that.do_substitute( c, n );
	} else assert( 0 );
}

void node::internalnode_full::print( std::ostream& out ) const {
	out << "internalnode_full:\n";
	internalnode::print( out );
	for ( uint8_t v = 0; v != number_characters; ++v ) {
		out << "map( \'" << idx2character( v ) << "\' ): " << do_map( idx2character( v ) ).cast_to_uint32() << '\n';
	}
}

node::nodeptr node::internalnode_full::do_map( character c ) const {
	uint16_t index = character2idx( c );
	return byte_io::read<nodeptr>( my_base() + index * 4 );
}
void node::internalnode_full::do_substitute( character c, nodeptr p ) {
	logfile() << boost::format( "internalnode_full[ %s ]::do_substitute( %s, %s )\n" ) % idx_ % char( c ) % p;
	if ( nodeptr prev = do_map( c ) ) {
		logfile() << boost::format( "previous inmate: %s\n" ) % prev.cast_to_uint32();
		logfile() << boost::format( "type: %s\n" ) % uint16_t( prev->type() );
	}

	assert( !do_map( c ) || !is_internalnode( do_map( c ) ) );
	uint16_t index = character2idx( c );
	unsigned char* data = my_base() + index * 4;
	return byte_io::write<nodeptr>( data, p );
}

void node::internalnode_half::print( std::ostream& out ) const {
	out << "internalnode_half\n";
	internalnode::print( out );
	bitstream in( my_base(), max_size * entry_size );
	character c;
	nodeptr p;
	for ( unsigned i = 0; i != max_size; ++i ) {
		in >> c >> p;
		out << "map( \'" << c << "\' ): " << p.cast_to_uint32() << '\n';
	}
}

node::nodeptr node::internalnode_half::do_map( character c ) const {
	using namespace node;
	bitstream in( my_base(), max_size * entry_size );
	character t;
	nodeptr res;
	for ( unsigned i = 0; i != max_size; ++i ) {
		in >> t >> res;
		if ( t == c ) return res;
	}
	return nodeptr();
}

bool node::internalnode_half::do_can_add() const {
	bitstream in( my_base(), max_size * entry_size );
	character c;
	nodeptr p;
	for ( unsigned i = 0; i != max_size; ++i ) {
		in >> c >> p;
		if ( !c ) return true;
	}
	return false;
}

void node::internalnode_half::do_substitute( character c, nodeptr p ) {
	bitstream in( my_base(), max_size * entry_size );
	character c2;
	nodeptr p2;
	for ( unsigned i = 0; i != max_size; ++i ) {
		in >> c2 >> p2;
		if ( c2 == c || !c2 ) {
			in.putback( p2.cast_to_uint32() );
			in.putback( c2 );
			in << c << p;
			return;
		}
	}
	assert( 0 );
}

void node::branch::print( std::ostream& out ) const {
	node::base_type::print( out );
	out << "next: " << next() << '\n';
	out << "string: " << reinterpret_cast<const char*>( string().c_str() ) << '\n';
}

character_string node::branch::string() const { // pseudo-virtual
	switch ( type() ) {
		case nodetypes::nearb:
			return nearbranchptr::cast_from_uint32( idx_ )->do_string();
		case nodetypes::farb:
			return farbranchptr::cast_from_uint32( idx_ )->do_string();
		default:
			//nodeptr::cast_from_uint32( idx_ )->print( std::cerr );
			assert( 0 );
			return character_string();
	}
}

void node::nearbranch::print( std::ostream& out ) const {
	out << "nearbranch:\n";
	branch::print( out );
}

character_string node::nearbranch::do_string() const {
	bitstream in( my_base(), bits_for<character>() * max_string_size );
	std::string res;
	character t;
	while ( 1 ) {
		in >> t;
		//logfile() << boost::format( "Read %c (%x).\n" ) % t % unsigned( t );
		if ( !t ) return character_string( res.c_str() );
		res.push_back( t );
	}
	return character_string( res.c_str() );
}


void node::nearbranch::set_string( character_string str ) {
	//logfile() << boost::format( "node::branch::set_string( %s )\n" ) % str;
	assert( str.size() );
	assert( str.size() <= max_string_size );
	bitstream out( my_base(), bits_for<character>() * max_string_size );
	for ( const character* p = str.c_str(); *p; ++p ) out << *p;
	out << nul_character;
	assert( string() == str );
}

void node::farbranch::set_stringarray( const stringarray& a ) {
	all_strings = &a;
}

void node::farbranch::print( std::ostream& out ) const {
	out << "farbranch[" << idx_ << "]:\n";
	branch::print( out );
	out << boost::format( "\tstring_idx: %s\n" ) % string_idx();
	out << boost::format( "\tstring_start: %s\n" ) % string_start();
	out << boost::format( "\tstring_size: %s\n" ) % string_size();
}
character_string node::farbranch::do_string() const {
	std::string mother = std::string( all_strings->get( string_idx() ) );
	mother.assign( mother.c_str() + string_start() );
	mother.resize( string_size() );
	return character_string( mother.c_str() );
}

void node::farbranch::set_string( uint16_t idx, uint16_t start, uint16_t size ) {
	logfile() << boost::format( "node::farbranch[ %s ]::set_string( %s, %s, %s )\n" ) % idx_ % idx % start % size;
	set_string_idx( idx );
	set_string_start( start );
	set_string_size( size );
	this->print( logfile() );
}



void node::leaf::print( std::ostream& out ) const {
	out << "leaf:\n";
	out << "start: " << start() << '\n';
}


void node::print( node::nodeptr p, std::ostream& out ) {
	using namespace node;
	out << '[' << p.cast_to_uint32() << "] ";
	switch ( p->type() ) {
		case nodetypes::ifull: 
			assert_downcast<internalnode_fullptr>( p )->print( out );
			break;
		case nodetypes::ihalf:
			assert_downcast<internalnode_halfptr>( p )->print( out );
			break;
		case nodetypes::leaf:
			assert_downcast<leafptr>( p )->print( out );
			break;
		case nodetypes::nearb:
			assert_downcast<nearbranchptr>( p )->print( out );
			break;
		case nodetypes::farb:
			assert_downcast<farbranchptr>( p )->print( out );
			break;
	}
}

