#ifndef LPC_NODE_H1102530057_INCLUDE_GUARD_
#define LPC_NODE_H1102530057_INCLUDE_GUARD_

/* This file is part of indexlib.
 * Copyright (C) 2005 Lus Pedro Coelho <luis@luispedro.org>
 *
 * Indexlib is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation and available as file
 * GPL_V2 which is distributed along with indexlib.
 * 
 * Indexlib is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of this program with any edition of
 * the Qt library by Trolltech AS, Norway (or with modified versions
 * of Qt that use the same license as Qt), and distribute linked
 * combinations including the two.  You must obey the GNU General
 * Public License in all respects for all of the code used other than
 * Qt.  If you modify this file, you may extend this exception to
 * your version of the file, but you are not obligated to do so.  If
 * you do not wish to do so, delete this exception statement from
 * your version.
 */


#include "character.h"
#include "pointer.h"
#include "thing.h"
#include "bitio.h"
#include "manager.h"
#include <ostream>
#include <boost/static_assert.hpp>
#include <boost/type_traits.hpp>

#define START_NODE( name ) START_THING( name, base_type )


class stringarray;

namespace node {
	memory_manager* get_manager();
	void set_manager( memory_manager* );
	

	namespace nodetypes {
		enum type_t { empty, ifull, ihalf, nearb, farb, leaf };
	}
	struct base_type : protected thing< thing_manager<&get_manager> > {
		protected:
			base_type( uint32_t i )
				:thing< thing_manager<&get_manager> >( i ) 
				{
				}
			friend class pointer<base_type>;
			void print( std::ostream& out ) const {
				out << "parent: " << parent() << '\n';
			}
		public:
			MEMBER( uint8_t, type, 0 )
			MEMBER( uint32_t, parent, 1 )
			void set_parent( const pointer<base_type>& p ) {
				set_parent( p.cast_to_uint32() );
			}

	};

	typedef pointer<base_type> nodeptr;
} // namespace

namespace byte_io {
	template<>
	inline
	node::nodeptr read<node::nodeptr>( const unsigned char* in )
	{
		return node::nodeptr::cast_from_uint32( read<uint32_t>( in ) );
	}

	template<>
	inline
	void write<node::nodeptr>( unsigned char* out, node::nodeptr d )
	{
		write<uint32_t>( out, d.cast_to_uint32() );
	}
}

namespace node {
	START_NODE( internalnode )
		protected:
			void print( std::ostream& out ) const {
				node::base_type::print( out );
				out << "count: " << count() << '\n';
				out << "start: " << start() << '\n';
				out << "hops: " << hops() << '\n';
			}
		public:
			MEMBER( uint16_t, count, 5 )
			MEMBER( uint16_t, start, 7 )
			MEMBER( uint16_t, hops, 9 )
			MEMBER( nodeptr, suflink, 11 )
			void assert_valid() const {
				assert( type() == nodetypes::ifull || type() == nodetypes::ihalf );
			}
		public:
			nodeptr map( character c ); // pseudo-virtual
			bool can_add() const;
			void add( character c, nodeptr n );
			void substitute( character c, nodeptr n );
	END_THING( internalnode )
	
	inline
	bool is_internalnode( nodeptr p ) {
		return p->type() == nodetypes::ihalf || p->type() == nodetypes::ifull;
	}

	START_THING( internalnode_full, internalnode )
		private:
			friend class internalnode;
			MY_BASE( 15 )
		public:
			static const bool is_concrete = true;
			static unsigned byte_size() { return 128; }
			static nodetypes::type_t node_type() { return nodetypes::ifull; }
			void print( std::ostream& out ) const;
			void assert_valid() const {
				assert( type() == node_type() );
			}
		private:
			nodeptr do_map( character c ) const;
			void do_substitute( character c, nodeptr p );
	END_THING( internalnode_full )

	START_THING( internalnode_half, internalnode )
		friend class internalnode;
		public:
			static const bool is_concrete = true;
			static unsigned byte_size() { return 64; }
			static nodetypes::type_t node_type() { return nodetypes::ihalf; }
			void print( std::ostream& out ) const;
			void assert_valid() const {
				assert( type() == node_type() );
			}
		private:
			MY_BASE( 15 )
			static const unsigned max_size = 8;
			static const unsigned entry_size = 37;
			nodeptr do_map( character c ) const;
			bool do_can_add() const;
			void do_substitute( character c, nodeptr p );
	END_THING( internalnode_half )

	START_NODE( branch )
		MEMBER( nodeptr, next, 5 )
		public:
			static const bool is_concrete = false;

			character_string string() const;
			void print( std::ostream& out ) const;
			void assert_valid() const {
				assert( type() == nodetypes::nearb || type() == nodetypes::farb );
			}
	END_THING( branch )

	inline
	bool is_branch( nodeptr p ) {
		return p->type() == nodetypes::nearb || p->type() == nodetypes::farb;
	}

	START_THING( nearbranch, branch )
		private:
			MY_BASE( 9 )
			friend class branch;
			character_string do_string() const;
		public:
			static const bool is_concrete = true;
			static unsigned byte_size() { return 16; }
			static nodetypes::type_t node_type() { return nodetypes::nearb; }

			static const unsigned max_string_size = 10;
			
			void set_string( character_string str );
			void print( std::ostream& out ) const;
			void assert_valid() const {
				assert( type() == node_type() );
			}
	END_THING( nearbranch )

	START_THING( farbranch, branch )
		public:
			MEMBER( uint16_t, string_idx,9 )
			MEMBER( uint16_t, string_start, 11 )
			MEMBER( uint16_t, string_size, 13 )
		private:
			friend class branch;
			character_string do_string() const;
			static const stringarray* all_strings;
		public:
			static const bool is_concrete = true;
			static unsigned byte_size() { return 16; }
			static nodetypes::type_t node_type() { return nodetypes::farb; }

			static void set_stringarray( const stringarray& );

			void set_string( uint16_t idx, uint16_t start, uint16_t len );
			void print( std::ostream& out ) const;
			void assert_valid() const {
				assert( type() == node_type() );
			}
	END_THING( farbranch )

	START_NODE( leaf )
		MEMBER( uint16_t, start, 5 )
		public:
			static const bool is_concrete = true;
			static unsigned byte_size() { return 16; }
			static nodetypes::type_t node_type() { return nodetypes::leaf; }
			void print( std::ostream& out ) const;
			void assert_valid() const {
				assert( type() == node_type() );
			}
	END_THING( leaf )

	template <typename D, typename B>
	D assert_downcast( const pointer<B>& base ) {
		BOOST_STATIC_ASSERT( (boost::is_convertible<typename D::value_type, B>::value ) );
		D res = D::cast_from_uint32( base.cast_to_uint32() );
		res->assert_valid();
		return res;
	}

	template <typename NodeType>
		void construct( void* data ) {
			BOOST_STATIC_ASSERT( NodeType::is_concrete );
			bzero( data, NodeType::byte_size() );
			*reinterpret_cast<unsigned char*>( data ) = NodeType::node_type();
		}

	void print( nodeptr p, std::ostream& out );

	struct node_mem_traits {
		typedef base_type value_type;
		typedef nodeptr pointer;
		static unsigned type_offset() { return 1; }
		static bool is_free( pointer p ) { return p->type() == nodetypes::empty; }
		static void mark_free( pointer p ) { p->set_type( nodetypes::empty ); }
		static unsigned size_of( pointer p ) {
			switch ( p->type() ) {
				case nodetypes::ifull: 	return internalnode_full::byte_size();
				case nodetypes::ihalf: 	return internalnode_half::byte_size();
				case nodetypes::farb: 	return farbranch	::byte_size();
				case nodetypes::nearb: 	return nearbranch	::byte_size();
				case nodetypes::leaf: 	return leaf		::byte_size();
				default: assert( 0 );
			}
		}

		static void print( std::ostream& out, nodeptr p ) { ::node::print( p, out ); }

		static unsigned min_size() { return 8; }
		static unsigned max_size() { return 128; }
		static void set_manager( memory_manager* m ) { ::node::set_manager( m ); }
	};
} // namespace node

#endif /* LPC_NODE_H1102530057_INCLUDE_GUARD_ */
