
/* This file is part of indexlib.
 * Copyright (C) 2005 Lus Pedro Coelho <luis@luispedro.org>
 *
 * Indexlib is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation and available as file
 * GPL_V2 which is distributed along with indexlib.
 * 
 * Indexlib is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * In addition, as a special exception, the copyright holders give
 * permission to link the code of this program with any edition of
 * the Qt library by Trolltech AS, Norway (or with modified versions
 * of Qt that use the same license as Qt), and distribute linked
 * combinations including the two.  You must obey the GNU General
 * Public License in all respects for all of the code used other than
 * Qt.  If you modify this file, you may extend this exception to
 * your version of the file, but you are not obligated to do so.  If
 * you do not wish to do so, delete this exception statement from
 * your version.
 */

#include "tree.h"
#include "logfile.h"
#include "mmap_manager.h"
#include "path.h"
#include <fstream>
#include <map>

using namespace node;

tree_result::tree_result( nodeptr p ) {
	dfs_fill( p );
	std::sort( results_.begin(), results_.end() );
	results_.erase( std::unique( results_.begin(), results_.end() ), results_.end() );
}

void tree_result::dfs_fill( nodeptr p ) {
	logfile() << boost::format( "dfs_fill( %s )\n" ) % p.cast_to_uint32();
	if ( is_internalnode( p ) ) {
		for ( uint8_t v = 0; v != number_characters; ++v ) {
			nodeptr next = assert_downcast<internalnodeptr>( p )->map( idx2character( v ) );
			if ( next ) dfs_fill( next );
		}
	} else if ( is_branch( p ) ) {
		dfs_fill( assert_downcast<branchptr>( p )->next() );
	} else if ( p->type() == nodetypes::leaf ) {
		leaf_dataptr data = leaf_dataptr::cast_from_uint32( assert_downcast<leafptr>( p )->start() );
		for ( leaf_data::iterator first = data->begin(), past = data->end(); first != past; ++first ) {
			logfile() << boost::format( "adding %s from %s\n" ) % *first % assert_downcast<leafptr>( p )->start();
			results_.push_back( *first );
		}
	}
}

std::vector<unsigned> tree_result::list() const {
	return results_;
}

unsigned tree_result::size() const {
	return results_.size();
}

bool tree_result::empty() const {
	return !size();
}


tree::tree( std::string base )
	:nodepool_( std::auto_ptr<memory_manager>( new mmap_manager( path_concat( base, "nodes" ).c_str() ) ) ),
	 leafdatapool_( std::auto_ptr<memory_manager>( new mmap_manager( path_concat( base, "leafs" ).c_str() ) ) ),
	 docnames_( path_concat( base, "docs" ) ),
	 rawstrings_( path_concat( base, "raws" ) ),
	 translatedstrings_( path_concat( base, "translated" ) ),
	 filebase_( base )
{
	std::ifstream in( path_concat( base, "tree" ).c_str() );
	if ( !in ) {
		root_ = allocate<internalnode_full>();
	} else {
		uint32_t root;
		in >> root;
		root_ = internalnode_fullptr::cast_from_uint32( root );
	}
	farbranch::set_stringarray( translatedstrings_ );
}

tree::~tree() {
	std::ofstream out( path_concat( filebase_, "tree" ).c_str() );
	out << root_.cast_to_uint32() << std::endl;
}

struct nodeptr_comparator {
	public:
		bool operator()( nodeptr a, nodeptr b ) const {
			return a.cast_to_uint32() < b.cast_to_uint32();
		}
};

struct tree::inserter {
	public:
		inserter( tree* m ):
			mother( m ),
		itranslated_( 0 )
		{ }

		void set_itranslated( unsigned t ) { itranslated_ = t; }

		void add( character_string, unsigned );
		nodeptr build_tail( const character*, unsigned );
	private:
		internalnodeptr break_up( nodeptr, const character* );
		tree* mother;
		const character* base_str_;
		unsigned itranslated_;
		unsigned docidx_;
};


void tree::add( const std::string original, std::string docname ) {
	const unsigned res = docnames_.add( docname );
	const unsigned iraw = rawstrings_.add( original );
	const character_string translated( original );
	const unsigned itranslated  = translatedstrings_.add( translated.cchar_str() );
	assert( itranslated == iraw );
	inserter ins( this );
	ins.set_itranslated( itranslated );
	ins.add( translated, res );
}

void tree::inserter::add( character_string cstr, const unsigned docidx ) {
	docidx_ = docidx;
	base_str_ = cstr.c_str();
	for ( const character* str = base_str_; *str; ++ str ) {
		nodeptr diverge;
		const character* divergechar;
		boost::tie( diverge, divergechar ) = mother->dive( str );
		if ( !*divergechar ) {
			continue;
		} else if ( diverge->type() == nodetypes::leaf ) {
			continue;
		} else {
			unsigned common = ( is_branch( diverge ) ?
					  common_chars( assert_downcast<branchptr>( diverge )->string().c_str(), str )
					  : 0 );
			const unsigned steps = divergechar - str + common;
			internalnodeptr previous = break_up( diverge, divergechar );
			for ( unsigned j = 0; j != steps; ++j ) {
				// walk the tree.from xa to a
				if ( common ) --common; // eat x
				if ( is_branch( diverge ) ) {
					diverge = nodeptr::cast_from_uint32( diverge->parent() );
					previous = assert_downcast<internalnodeptr>( diverge );
				}
				if ( is_internalnode( diverge ) && assert_downcast<internalnodeptr>( diverge )->suflink() ) {
					diverge = assert_downcast<internalnodeptr>( diverge )->suflink();
				}
				while ( common ) {
					diverge = assert_downcast<internalnodeptr>( diverge )->map( *( str + steps - common ) );
					--common;
					branchptr branch = assert_downcast<branchptr>( diverge );
					if ( branch->string().size() > common ) break;
					else common -= branch->string().size();
				}
				internalnodeptr next = break_up( diverge, divergechar + j );
				previous->set_suflink( next );
				previous = next;
			}
			str += steps - 1;
		}
	}
}

internalnodeptr tree::inserter::break_up( nodeptr diverge, const character* divergechar ) {
	if ( is_internalnode( diverge ) ) {
		internalnode_fullptr idiverge = assert_downcast<internalnode_fullptr>( diverge );
		nodeptr tail = build_tail( divergechar + 1, docidx_ );
		logfile() << boost::format( "Setting parent: %s(%s) for %s\n" ) % idiverge % diverge % tail;
		tail->set_parent( idiverge );
		idiverge->add( *divergechar, tail );
		return idiverge;
	}
	if ( is_branch( diverge ) ) {
		branchptr top = assert_downcast<branchptr>( diverge );
		// If it's a branch, dive() returns the start of the branch and not the
		// diverging character!
		// Needs to break the branch:
		const unsigned before = common_chars( top->string().c_str(), divergechar );
		const int after = top->string().size() - before - 1;

		logfile() << boost::format( "top: %s\n" "before: %s\nafter: %s\n" ) % top % before % after;
		logfile() << boost::format( "top->string(): %s\n" ) % reinterpret_cast<const char*>( top->string().c_str() );

		if ( before == char_len( divergechar ) ) return internalnodeptr::cast_from_uint32( 0 );

		internalnode_fullptr middle = mother->allocate<internalnode_full>();

		if ( before ) {
			if ( after > 0 ) {
				branchptr bottom = ( unsigned( after ) > nearbranch::max_string_size ?
						branchptr( mother->allocate<node::farbranch>() ):
						branchptr( mother->allocate<node::nearbranch>() ) );
				logfile() << boost::format( "after : %d\n" ) % after;
				bottom->set_parent( middle );
				bottom->set_next( top->next() );
				if ( after > nearbranch::max_string_size ) {
					farbranchptr topf = assert_downcast<farbranchptr>( top );
					assert_downcast<farbranchptr>( bottom )->
						set_string( topf->string_idx(), topf->string_start() + before + 1, after );
				} else {
					assert_downcast<nearbranchptr>( bottom )->
						set_string( right( top->string(), after ) );
				}
				middle->add( top->string()[ before ], bottom );
			}
			top->set_next( middle );
			if ( top->type() == nodetypes::nearb ) {
				assert_downcast<nearbranchptr>( top )->set_string( left( top->string(), before ) );
			} else {
				farbranchptr topf = assert_downcast<farbranchptr>( top );
				topf->set_string( topf->string_idx(), topf->string_start(), before );
			}
			middle->set_parent( top );
		} else {
			switch ( nodeptr::cast_from_uint32( top->parent() )->type() ) {
				case nodetypes::ifull: {
					internalnode_fullptr p = internalnode_fullptr::cast_from_uint32( top->parent() );
					//assert( p->map( divergechar[ -1 ] ) == top );
					p->substitute( divergechar[ -1 ], middle );
					middle->set_parent( p );
					if ( after ) {
						middle->add( top->string()[ 0 ], top );
						if ( top->type() == nodetypes::nearb ) {
							assert_downcast<nearbranchptr>( top )->set_string( top->string().c_str() + 1 );
						} else {
							farbranchptr topf = assert_downcast<farbranchptr>( top );
							topf->set_string( topf->string_idx(), topf->string_start() + 1, topf->string_size() - 1 );
						}
						top->set_parent( middle );
					} else {
						middle->add( top->string()[ 0 ], top->next() );
						mother->deallocate( top );
					}
					break;
				}
				case nodetypes::farb:
				case nodetypes::nearb: {
					branchptr p = branchptr::cast_from_uint32( top->parent() );
					p->set_next( middle );
					break;
				}
				default: {
					logfile() << boost::format( "Going to crash: [%s] has type: %d\n" ) 
						% top->parent()
						% unsigned( nodeptr::cast_from_uint32( top->parent() )->type() );
					assert( 0 );
				}
			}
		}

		if ( after > 0 ) { // Is this correct?
			nodeptr tail = build_tail( divergechar + before + 1, docidx_ );
			middle->add( divergechar[ before ], tail );
			tail->set_parent( middle );
		}
		return middle;
	}
	assert( 0 );
	return internalnodeptr::cast_from_uint32( 0 );
}

std::auto_ptr<indexlib::result> tree::search( const std::string str ) const {
	nodeptr diverge;
	const character* divergechar;
	const character_string trans = str;

	boost::tie( diverge, divergechar ) = dive( root_, trans.c_str() );


	if ( !*divergechar ||
			( is_branch( diverge )
			&& common_chars( divergechar, assert_downcast<branchptr>( diverge )->string().c_str() ) == char_len( divergechar ) ) ) { // FOUND
		return std::auto_ptr<indexlib::result>( new tree_result( diverge ) );
	} else {
		logfile() << boost::format( "tree::search( %s ) found NOTHING\n" ) % str;
		return std::auto_ptr<indexlib::result>( new tree_result() );
	}
}
	

template <typename Node>
pointer<Node> tree::allocate() {
	nodeptr res = nodepool_.allocate( Node::byte_size() );
	construct<Node>( res.raw_pointer() );
	return assert_downcast< pointer<Node> >( res );
}


boost::tuple<nodeptr, const character*> tree::dive( nodeptr start, const character* str ) const {
	logfile() << boost::format( "tree::dive() for %s\n" ) % reinterpret_cast<const char*>( str );
	while ( *str ) {
		logfile() << boost::format( "tree::dive() looking at %s\n" ) % reinterpret_cast<const char*>( str );
		if ( is_internalnode( start ) ) {
			nodeptr next = assert_downcast<internalnodeptr>( start )->map( *str );
			logfile() << boost::format( "assert_downcast<internalnodeptr>( start )->map(%s) = " ) % 
				*reinterpret_cast<const char*>( str ) << next << '\n';
			if ( !next ) goto end;
			else {
				start = next;
				++str;
			}
		} else if ( is_branch( start ) ) {
			branchptr branch = assert_downcast<branchptr>( start );
			unsigned common = common_chars( str, branch->string().c_str() );
			logfile() << boost::format( 
					"\tbranch->string() == %s\n"
					"\tcommon = %s\n"
					"\tchar_len( branch->string() ) = %s\n" )
				% branch->string().c_str()
				% common
				% char_len( branch->string().c_str() );

			if ( common == char_len( branch->string().c_str() ) ) {
				str += common;
				start = branch->next();
			} else goto end;
		} else if ( start->type() == nodetypes::leaf ) {
			goto end;
		} else {
			assert( 0 );
		}
	}
end:



	logfile() << boost::format( "tree::dive()::end: *str = [%s]\n" ) % *reinterpret_cast<const char*>( str );
	return boost::make_tuple( start, str );
}

nodeptr tree::inserter::build_tail( const character* str, unsigned ref ) {
	leafptr leafp = mother->allocate<leaf>();
	leaf_dataptr dp = mother->leafdatapool_.allocate( leaf_data::start_bytes() );
	leaf_data::construct( dp.raw_pointer() );
	leafp->set_start( dp.cast_to_uint32() );
	dp->add_reference( ref );

	if ( !*str ) return leafp;

	branchptr res;
	if ( char_len( str ) <= nearbranch::max_string_size ) {
		nearbranchptr n = mother->allocate<nearbranch>();
		n->set_string( str );
		res = n;
	} else {
		farbranchptr f = mother->allocate<farbranch>();
		f->set_string( itranslated_, str - base_str_, char_len( str ) );
		res = f;
	}

	leafp->set_parent( res );
	res->set_next( leafp );
	return res;
}

